# Class for classical propagation methods
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import axes
from tqdm.auto import trange
import qutip as qt
from typing import Callable
from .utils import *


class DynamicsCS:
    """
        Class for closed-system dynamics. **All input parameters must be in
        atomic units to ensure consistency. Please be sure to convert your
        parameters to atomic units prior to instantiation.**

    """

    def __init__(self, n_basis: int = 128, xo: float = 1.0, po: float = 0.0, mass: float = 1.0,
                 omega: float = 1.0) -> None:
        """
        Args:
            n_basis (int): Number of states to include in the chosen representation. If basis
                = 'ladder', this is the Fock cutoff and defines the number of states
                used for representing the ladder operators. If basis = 'coordinate',
                this defines the number of points for the position and momenta.

            xo (float, optional): Defines the displacement of the initial state in the position
                coordinate. Default is 1.0 Bohr.

            po (float, optional): Defines the displacement of the initial state in the position
                coordinate. Default is 1.0 au.

            mass (float, optional): Defines the mass of the particle/system of interest.
                Default is 1.0 au.

            omega (float, optional): Frequency of harmonic oscillator.
                Default is 1.0 au.

        """
        #--------- Required Attributes Populated During Execution ----------#
        self.n_basis                   = n_basis
        self.xo                        = xo
        self.po                        = po
        self.mass                      = mass
        self.hbar                      = 1.0
        self.omega                     = omega
        #--------- Below are Attributes Populated During Execution ---------#
        self.total_time                = 0.
        self.n_tsteps                  = 0.
        self._KE_op                    = None
        self._PE_op                    = None
        self.H_op                      = None
        self.prop_KE_op                = None
        self.prop_PE_op                = None
        self.prop_H_op                 = None
        # Grid operators
        self._KE_grid                  = None
        self._PE_grid                  = None
        self.H_grid                    = None
        self.PE_prop_grid              = None
        self.KE_prop_grid              = None


    def _get_xgrid(self, x_min: float, x_max: float) -> None:
        """
        Populate the `self.x_grid` and `self.dx` attributes. This function
        generates an array of `self.n_basis` evenly spaced values between
        `x_min` and `x_max`.

        Args:
            x_min (float): Minimum value of x-coordinates
            x_max (float): Maximum value of x-coordinates

        Returns:
            self.dx (float): Spacing between points in the x-coordinate grid.
            self.xgrid (array_like): Array of grid points from x_min to x_max with spacing of dx
        """
        dx = (x_max - x_min) / self.n_basis
        x_grid = np.arange(-self.n_basis / 2, self.n_basis / 2) * dx
        self.dx = dx
        self.x_grid = x_grid
        return


    def _get_pgrid(self, x_min: float, x_max: float, reorder: bool = True) -> None:
        """
        Populate the `self.p_grid` and `self.dp` attributes. This function
        generates an array of `self.n_basis` evenly spaced values.

        Args:
            x_min (float): Minimum value of x-coordinates
            x_max (float): Maximum value of x-coordinates
            reorder (bool): Boolean flag to determine whether points should be reordered to be
                compatible with the FFT routine or not.

        Returns:
            self.dp (float): Spacing between points in the p-coordinate grid.
            self.pgrid (array_like): Array of momentum grid points
        """
        dp = 2 * np.pi / (x_max - x_min)
        pmin = -dp * self.n_basis / 2
        pmax = dp * self.n_basis / 2
        plus_pgrid = np.linspace(0, pmax, self.n_basis // 2 + 1)
        minus_pgrid = - np.flip(np.copy(plus_pgrid))
        if reorder:
            p_grid = np.concatenate((plus_pgrid[:-1], minus_pgrid[:-1]))
        else:
            p_grid = np.concatenate((minus_pgrid, plus_pgrid))
        self.p_grid = p_grid
        self.dp = dp
        return


    def set_coordinate_operators(self, x_min: float = -7., x_max: float = 7., reorder_p: bool = True) -> None:
        """
        Populate the `self.x_grid`, `self.p_grid`, `self.dx`, and `self.dp`
        attributes. This functions generates an array of `self.n_basis`
        evenly spaced values.

        Args:
            x_min : float
                Minimum value of x-coordinates
            x_max : float
                Maximum value of x-coordinates
            reorder_p : bool
                Boolean flag to determine whether momentum values should be
                reordered to be compatible with the FFT routine or not.

        Returns:
            self.dx : float
                Spacing between points in the x-coordinate grid.
            self.xgrid : array_like
                Array of x-values
            self.dp : float
                Spacing between points in the p-coordinate grid.
            self.pgrid : array_like
                Array of p-values
        """
        self._get_xgrid(x_min, x_max)
        self._get_pgrid(x_min, x_max, reorder=reorder_p)
        return


    def initialize_operators(self):
        """
            Function to initialize core operators in the chosen basis.

        """

        self.a_op = qt.destroy(self.n_basis)
        self.x_op = qt.position(self.n_basis)
        self.p_op = qt.momentum(self.n_basis)
        return


    def _set_hamiltonian_grid(self, potential_type: str = 'harmonic', **kwargs):
        if potential_type == 'harmonic':

            # Set attributes for the coordinate basis
            self._PE_grid = self.mass * self.omega ** 2 * self.x_grid ** 2 / 2.
            self._KE_grid = self.p_grid ** 2 / (2. * self.mass)
            self.H_grid = self._PE_grid + self._KE_grid
        elif potential_type == 'quartic':
            if kwargs:
                a0 = 0.0
                a1 = 0.0
                a2 = 0.0
                a3 = 0.0
                a4 = 0.0
                if 'a0' in kwargs:
                    a0 = kwargs['a0']
                if 'a1' in kwargs:
                    a1 = kwargs['a1']
                if 'a2' in kwargs:
                    a2 = kwargs['a2']
                if 'a3' in kwargs:
                    a3 = kwargs['a3']
                if 'a4' in kwargs:
                    a4 = kwargs['a4']
                if 'x0' in kwargs:
                    x0 = kwargs['x0']
                # Assume that all inputs have the proper atomic units:
                cf = 1.0
                xi = self.x_op

            else:
                # Define relevant parameters
                cf = convert_eV_to_au(1.)
                x0 = 1.9592
                a0 = 0.0
                a1 = 0.429
                a2 = -1.126
                a3 = -0.143
                a4 = 0.563
                # Do calculation for ladder basis
                xi = self.x_grid / x0
            self._PE_grid = cf * (a0 + a1 * xi + a2 * xi ** 2 + a3 * xi ** 3 + a4 * xi ** 4)
            self._KE_grid = self.p_grid ** 2 / (2. * self.mass)
            self.H_grid = self._PE_grid + self._KE_grid
        return


    def _set_hamiltonian_qt(self, potential_type: str = 'harmonic', **kwargs):
        if potential_type == 'harmonic':
            # Set attributes for the ladder basis
            self.H_op = self.hbar * self.omega * (self.a_op.dag() * self.a_op + 0.5)
            self._KE_op = self.p_op ** 2 / (2. * self.mass)
            self._PE_op = self.mass * self.omega ** 2 * self.x_op ** 2 / 0.5
            self.H_xp_op = self._PE_op + self._KE_op
        elif potential_type == 'quartic':
            if kwargs:
                a0 = 0.0
                a1 = 0.0
                a2 = 0.0
                a3 = 0.0
                a4 = 0.0
                if 'a0' in kwargs:
                    a0 = kwargs['a0']
                if 'a1' in kwargs:
                    a1 = kwargs['a1']
                if 'a2' in kwargs:
                    a2 = kwargs['a2']
                if 'a3' in kwargs:
                    a3 = kwargs['a3']
                if 'a4' in kwargs:
                    a4 = kwargs['a4']
                if 'x0' in kwargs:
                    x0 = kwargs['x0']
                # Assume that all inputs have the proper atomic units:
                cf = 1.0
                xi = self.x_op

            else:
                # Define relevant parameters
                cf = convert_eV_to_au(1.)
                x0 = 1.9592
                a0 = 0.0
                a1 = 0.429
                a2 = -1.126
                a3 = -0.143
                a4 = 0.563
                # Do calculation for ladder basis
                xi = self.x_op / x0
            self.x0 = x0
            self._PE_op = cf * (a0 + a1 * xi + a2 * xi ** 2 + a3 * xi ** 3 + a4 * xi ** 4)
            self._KE_op = self.p_op ** 2 / (2. * self.mass)
            self.H_op = self._PE_op + self._KE_op
            return

    def set_hamiltonian(self, potential_type: str = 'harmonic', **kwargs):
        """
        Function to define Hamiltonian.

        Args:
            potential_type : str
                String defining the type of potential energy surface.
                Available options are: ('harmonic', 'quartic', ...)

                Note: You can manually define your potential energy using the functions:
                    - set_H_grid_with_custom_potential
                    - set_H_op_with_custom_potential

        """

        if potential_type == 'harmonic':
            self._set_hamiltonian_grid(potential_type=potential_type, **kwargs)
            self._set_hamiltonian_qt(potential_type=potential_type, **kwargs)
        elif potential_type == 'quartic':
            self._set_hamiltonian_grid(potential_type=potential_type, **kwargs)
            self._set_hamiltonian_qt(potential_type=potential_type, **kwargs)
        else:
            print('Error, this potential type has not yet been implemented!')
            print('Set your parameters with the custom functions!')
        return


    def set_H_grid_with_custom_potential(self, custom_function: Callable, **kwargs):
        """
        Function to allow for user-defined potential defined by custom_function. Must be a function of qutip operators.

        Args:
            custom_function (Callable): Function that defines the custom potential
                energy. Must return an array

        """
        potential = custom_function(**kwargs)
        self._PE_grid = potential
        self._KE_grid = self.p_grid ** 2 / (2. * self.mass)
        self.H_grid = self._PE_grid + self._KE_grid
        return


    def set_H_op_with_custom_potential(self, custom_function: Callable, **kwargs):
        """
        Function to allow for user-defined potential defined by custom_function. Must be a function of qutip operators.

        Args:
            custom_function (Callable): Function that defines the potential
                energy in terms of qutip QObj operators. Must return a qutip.Qobj
        """
        potential = custom_function(**kwargs)
        self._PE_op = potential
        self._KE_op = self.p_op ** 2 / (2. * self.mass)
        self.H_op = self._PE_op + self._KE_op
        return


    def set_initial_state(self, wfn_omega: float = 1.0):
        """
        Function to define the initial state. By default, a coherent state is
        used as the initial state defined in the basis chosen upon instantiation

        Args:
            wfn_omega (float, optional): Defines the frequency/width of the initial state.
                Default is 1.0 au.
        """

        alpha_val = (self.xo + 1j * self.po) / np.sqrt(2)
        psio = qt.coherent(self.n_basis, alpha=alpha_val)
        # Now populate the initial state in the grid basis
        normalization = (self.mass * wfn_omega / np.pi / self.hbar) ** (0.25)
        exponential = np.exp(-1 * (self.mass * wfn_omega / self.hbar / 2) *
                             ((self.x_grid - self.xo) ** 2)
                             + 1j * self.po * self.x_grid / self.hbar
                             )

        coherent_state = normalization * exponential
        # Set the attributes
        self.psio_grid = coherent_state
        self.psio_op = psio
        return



    def custom_grid_state_initialization(self, function_name: Callable, **kwargs):
        """
        Function to allow for customized grid state initialization.

        Args:
            function_name (Callable): name of user-defined function that returns
                the initial state. Must return an array
        """

        self.psio_grid = function_name(**kwargs)
        return

    def custom_ladder_state_initialization(self, function_name: Callable, **kwargs):
        """
        Function to allow for customized ladder state initialization.

        Args:
            function_name (Callable): name of user-defined function that returns
                the initial state. Must return a qutip.Qobj.
        """

        self.psio_op = function_name(**kwargs)
        return

    def set_propagation_time(self, total_time: float, n_tsteps: int):
        """
        Function to define the propagation time, an array of times from
        t=0 to total_time, with n_tsteps equally-spaced steps.

        Args:
        total_time : float
            The total time for which we wish to compute the dynamics.
        n_tsteps : int
            The number of equally-spaced time steps used to compute the dynamics

        Returns:
        self.tlist : array-like

        """

        self.tlist = np.linspace(0., total_time, n_tsteps+1)
        self.dt = self.tlist[1] - self.tlist[0]
        self.n_tsteps = n_tsteps
        return


    def propagate_qt(self, solver_options : dict = None):
        """
        Function used to propagate with qutip.

        Args:
            solver_options (dict): A dictionary of arguments to pass to the qutip.sesolve function

        Returns:
            dynamics_results (array-like): array containing the propagated state

        """

        options = {'nsteps': len(self.tlist),
                    'progress_bar': True}

        if solver_options:
            for key in solver_options:
                options[key] = solver_options[key]

        results = qt.sesolve(self.H_op, self.psio_op, self.tlist,
                             options=options)

        self.dynamics_results_op = results.states
        return


    def propagate_SOFT(self):
        """
        Function used to propagate with the 2nd-Order Trotter Expansion.

        $$
        e^{- \\frac{i}{\\hbar} H t} \\approx e^{- \\frac{i}{\\hbar} V t/2} e^{- \\frac{i}{\\hbar} T t} e^{- \\frac{i}{\\hbar} V t/2} + \\mathcal{O}^{3}
        $$

        Returns:
            dynamics_results_grid (array-like): array containing the propagated state
                                                shape (n_tsteps x self.n_basis)

        """
        self.tau = self.tlist[1] - self.tlist[0]
        PE_prop = np.exp(-1.0j * self._PE_grid / 2 * self.tau / self.hbar)
        KE_prop = np.exp(-1.0j * self._KE_grid * self.tau / self.hbar)

        self.PE_prop_grid = PE_prop
        self.KE_prop_grid = KE_prop

        propagated_states = [self.psio_grid]
        psi_t = self.psio_grid
        for ii in range(1, len(self.tlist)):
            psi_t_position_grid = PE_prop * psi_t
            psi_t_momentum_grid = KE_prop * np.fft.fft(psi_t_position_grid, norm="ortho")
            psi_t = PE_prop * np.fft.ifft(psi_t_momentum_grid, norm="ortho")
            propagated_states.append(psi_t)

        self.dynamics_results_grid = np.asarray(propagated_states)
        return



import numpy as np

## Spin-Boson Model parameters 
GAMMA_DA = 1 # diabatic coupling
EPSILON = 1
BETA = 5 # inverse finite temperature beta = 1 / (k_B * T)
XI = 0.1
OMEGA_C = 2

print("SPIN-BOSON Model parameter")
print("        epsilon =", EPSILON)
print("             xi =", XI)
print("        omega_c =", OMEGA_C)

# Spin-up and spin-down states
spin_up = np.array([1.0, 0.0], dtype=np.float64)
spin_down = np.array([0.0, 1.0], dtype=np.float64)

## General Constants for simulation
TIME_STEPS = 500          # number of time steps
au2ps = 0.00002418884254   # Conversion of attoseconds to atomic units
timeau = 12.409275
DT = 20 * au2ps * timeau  # time step in au

FINAL_TIME = TIME_STEPS * DT # final time
DOF_E = 2 # number of electronic states
DOF_E_SQ = DOF_E * DOF_E

##Simulation Parameter for TT-TFD
# TFD parameter: for Discretized nuclear DOFs
DOF_N = 50 # number of nuclear DOF
OMEGA_MAX = 10

# TT constants
eps = 1e-12           # tt approx error
dim = DOF_N         # number of coords
occ = 10                 # maximum occupation number; low for harmonic systems
MAX_TT_RANK = 10

print("      omega_max =", OMEGA_MAX)
print("     time steps =", TIME_STEPS)
print("             DT =", DT)
print("     final time =", FINAL_TIME)
print("          DOF_E =", DOF_E)
print("          DOF_N =", DOF_N)

##Simulation Parameter for GQME
MEM_TIME = DT * TIME_STEPS
HBAR = 1
MAX_ITERS = 30
CONVERGENCE_PARAM = 10.**(-10.)

##Parameter for output files
PARAM_STR = "_Spin-Boson_Ohmic_TT-TFD_b%sG%s_e%s_"%(BETA, GAMMA_DA, EPSILON)
PARAM_STR += "xi%swc%s_wmax%s_dofn%s"%(XI, OMEGA_C, OMEGA_MAX, DOF_N)

# Pauli matrices
X = np.array([[0, 1], [1, 0]], dtype=np.complex128)
Y = np.array([[0, -1j], [1j, 0]], dtype=np.complex128)
Z = np.array([[1, 0], [0, -1]], dtype=np.complex128)
I = np.eye(2, dtype=np.complex128)




import numpy as np
import scipy.linalg as LA
from typing import List, Tuple, Optional, Dict, Any, Union

from qiskit import transpile
from qiskit_aer import AerSimulator
#from qiskit.primitives import Estimator
from qiskit_ibm_runtime import EstimatorV2 as Estimator
from qiskit.quantum_info import SparsePauliOp
from qiskit.circuit import QuantumCircuit

from . import trans_basis as tb
from . import dilation_circuit as dc
from .numerical_methods import DynamicsOS


def expand(Gmat_org: np.ndarray, Norg: int, Nexpand: int) -> np.ndarray:
    """
    Expand the propagator in the vectorized density matrix representation.

    This function expands the original propagator matrix to match the dimensions
    of the vectorized density matrix representation (e.g., match the 2^N dimension).

    Args:
        Gmat_org (np.ndarray): Original propagator matrix.
        Norg (int): Original system dimension.
        Nexpand (int): Expanded system dimension.

    Returns:
        np.ndarray: The expanded propagator matrix with shape (Nexpand**2, Nexpand**2).
    """
    Gnew = np.zeros((Nexpand**2, Nexpand**2), dtype=np.complex128)
    for i in range(Norg):
        for j in range(Norg):
            for k in range(Norg):
                for l in range(Norg):
                    Gnew[i * Nexpand + j, k * Nexpand + l] = Gmat_org[i * Norg + j, k * Norg + l]
    return Gnew


def gen_Kraus_list(Gmat: np.ndarray, N: int, tol: float = 1e-5) -> List[np.ndarray]:
    """
    Generate the Kraus operators from the propagator with a given tolerance.

    This function computes the Kraus operator representation by constructing the Choi
    matrix from the propagator matrix and performing an eigenvalue decomposition.

    Args:
        Gmat (np.ndarray): Matrix of the propagator with shape (N^2, N^2).
        N (int): The system Hilbert space dimension.
        tol (float, optional): Tolerance threshold for the Kraus operator representation.
            Operators with eigenvalues below tol are ignored. Defaults to 1e-5.

    Returns:
        List[np.ndarray]: A list of Kraus operators.
    """
    # Define the Choi matrix from the propagator matrix
    C_mat = np.zeros(Gmat.shape, dtype=np.complex128)
    for i in range(N):
        for j in range(N):
            C_matij = np.zeros(Gmat.shape, dtype=np.complex128)
            for k in range(N):
                for l in range(N):
                    C_matij[i * N + k, l * N + j] = Gmat[j * N + k, l * N + i]
            C_mat += C_matij

    Kraus: List[np.ndarray] = []
    eigenvalues, eigenvectors = LA.eigh(C_mat)
    for idx in range(len(eigenvalues)):
        if eigenvalues[idx] > tol:
            Mi = np.sqrt(eigenvalues[idx]) * eigenvectors[:, idx].reshape(N, N)
            Kraus.append(Mi.conj().T)
    return Kraus


class QubitDynamicsOS(DynamicsOS):
    """
    Class for simulating quantum dynamics using either vectorized density matrix or Kraus operator representations.

    This class provides methods to initialize state vectors, construct quantum circuits,
    and perform quantum simulations using Qiskit backends.
    """

    def __init__(self, rep: str = 'Density', **kwargs: Any) -> None:
        """
        Initialize a QubitDynamicsOS instance.

        Depending on the representation, either "Density" or "Kraus", the number of qubits is computed.
        Additional keyword arguments are passed to the base DynamicsOS class.

        Args:
            rep (str, optional): Representation type, either 'Density' or 'Kraus'. Defaults to 'Density'.
            **kwargs: Additional keyword arguments for the DynamicsOS initializer.
        """
        super().__init__(**kwargs)

        if rep == 'Density':
            # Vectorized density matrix representation
            self.rep: str = 'Density'
            self.Nqb: int = int(np.log2(self.Nsys**2))
        elif rep == 'Kraus':
            # Kraus operator representation
            self.rep = 'Kraus'
            self.Nqb = int(np.log2(self.Nsys))

        # The counting qubits bit string and observable matrix are initialized to None.
        self.count_str: Optional[List[str]] = None
        self.observable: Optional[np.ndarray] = None

        # Default dilation method for quantum simulation.
        self.dilation_method: str = 'Sz-Nagy'

    def set_dilation_method(self, method: str) -> None:
        """
        Set the dilation method for quantum simulation.

        Args:
            method (str): The dilation method, e.g., 'Sz-Nagy', 'SVD', or 'SVD-Walsh'.
        """
        self.dilation_method = method

    def set_count_str(self, count_str: List[str]) -> None:
        """
        Set the counting bit string for measurement.

        Args:
            count_str (List[str]): The counting bit string.
        """
        self.count_str = count_str

    def set_observable(self, observable: np.ndarray) -> None:
        """
        Set the observable for the quantum simulation.

        Args:
            observable (np.ndarray): The observable matrix.
        """
        self.observable = observable

    def init_statevec_vecdens(self) -> Tuple[np.ndarray, float]:
        """
        Initialize the state vector from the initial density operator using vectorized representation.

        The initial density matrix is reshaped into a vector and normalized.

        Returns:
            Tuple[np.ndarray, float]: A tuple containing the normalized state vector and the norm
            of the original vectorized density matrix.
        """
        vec_rho0 = self.rho0.reshape(self.Nsys**2)
        norm0 = LA.norm(vec_rho0, 2)
        statevec = vec_rho0 / norm0
        return statevec, norm0

    def init_statevec_Kraus(self, tol: float = 1e-6) -> Tuple[List[np.ndarray], List[float]]:
        """
        Initialize state vectors from the initial density operator using the Kraus operator representation.

        The density matrix is decomposed using eigenvalue decomposition, and eigenstates with eigenvalues
        below the specified tolerance are ignored.

        Args:
            tol (float, optional): Tolerance for ignoring eigenstates with small eigenvalues. Defaults to 1e-6.

        Returns:
            Tuple[List[np.ndarray], List[float]]: A tuple containing a list of state vectors and a list of
            corresponding probabilities.
        """
        eigenvalues, eigenvectors = LA.eigh(self.rho0)
        statevec: List[np.ndarray] = []
        prob: List[float] = []
        for i in range(len(eigenvalues) - 1, -1, -1):
            if abs(eigenvalues[i]) < tol:
                break
            prob.append(eigenvalues[i])
            statevec.append(eigenvectors[:, i])
        return statevec, prob

    def _get_qiskit_observable(self, Isdilate: bool = False, tol: float = 5e-3) -> SparsePauliOp:
        """
        Prepare and return the Qiskit observable operator.

        Converts the observable matrix to its Pauli representation and returns a SparsePauliOp.

        Args:
            Isdilate (bool, optional): Flag indicating whether to use the dilated observable.
                Defaults to False.
            tol (float, optional): Tolerance for the Pauli decomposition. Defaults to 5e-3.

        Returns:
            SparsePauliOp: The Qiskit representation of the observable.
        """
        if self.observable is None:
            print('Error: observable is None')

        if Isdilate:
            num_qubits = self.Nqb + 1
            Obs_mat = np.zeros((2 * self.Nsys, 2 * self.Nsys), dtype=np.complex128)
            Obs_mat[:self.Nsys, :self.Nsys] = self.observable[:self.Nsys, :self.Nsys]
        else:
            num_qubits = self.Nqb
            Obs_mat = self.observable

        Obs_paulis_dic = tb.ham_to_pauli(Obs_mat, num_qubits, tol=tol)

        # Prepare the Qiskit observable from the Pauli strings of the observable matrix.
        data: List[str] = []
        coef: List[float] = []
        for key in Obs_paulis_dic:
            data.append(key)
            coef.append(Obs_paulis_dic[key])
        obs_q = SparsePauliOp(data, coef)
        return obs_q

    def qc_simulation_kraus(
        self,
        time_arr: List[float],
        Kraus: Optional[Dict[int, List[np.ndarray]]] = None,
        backend: Any = AerSimulator(), 
        Gprop: Optional[List[np.ndarray]] = None,
        tolk: float = 1e-5,
        tolo: float = 1e-5,
        Is_store_circuit = False,
        **kwargs: Any
        ) -> np.ndarray:
        """
        Perform quantum simulation using the Kraus operator representation.

        This method simulates the quantum system dynamics over a series of time steps using a Kraus operator-based approach.
        It constructs quantum circuits for each Kraus operator and initial state, runs the simulation using Qiskit's Estimator,
        and accumulates the measurement results.

        Args:
            time_arr (List[float]): List of time steps for simulation.
            
            Kraus (Optional[Dict[int, List[np.ndarray]]], optional): Dictionary mapping time step index to a list of Kraus operators.
                If None, Kraus operators are generated from the propagator. Defaults to None.
            backend (Any, optional): Quantum simulation backend. Defaults to AerSimulator().
            Gprop (Optional[List[np.ndarray]], optional): Propagator matrix (or list of matrices) for simulation.
                If None, it will be calculated. Defaults to None.
            tolk (float, optional): Tolerance for generating Kraus operators. Defaults to 1e-5.
            tolo (float, optional): Tolerance for observable decomposition. Defaults to 1e-5.
            Is_store_circuit (bool, optional): If True, store the generated quantum circuits at each time step. Defaults to False.
            **kwargs: Additional keyword arguments for propagator calculation.

        Returns:
            Dict[str, Any]: A dictionary containing quantum simulation results, including:
                - 'data' (np.ndarray): Array of shape (nsteps,), containing the accumulated observable expectation values at each time step.
                - 'circuits' (List[List[QuantumCircuit]], optional): If `Is_store_circuit` is True, this field contains a list of lists of quantum circuits.
                  Each outer list corresponds to a time step, and each inner list contains all circuits used for that step (across different Kraus operators and initial states).
        """
        nsteps = len(time_arr)

        # Generate Kraus operators if not provided.
        if Kraus is None:
            Kraus = {}
            if Gprop is None:
                print('Calculating the propagator')
                Gprop = self.Gt_matrix_expo(time_arr, **kwargs)
            print('Generating the Kraus operators')
            for i in range(nsteps):
                print('At step', i, 'of', nsteps)
                Kraus[i] = gen_Kraus_list(Gprop[i], self.Nsys, tol=tolk)
        print('Kraus operator generation complete')

        # Perform Qiskit simulation using the Estimator.
        estimator = Estimator(mode=backend)

        statevec, prob = self.init_statevec_Kraus()
        n_inistate = len(statevec)
        print('Number of initial states in the density matrix:', n_inistate)
        print('Probabilities:', prob)

        # Obtain the Qiskit observable.
        obs_q = self._get_qiskit_observable(Isdilate=True, tol=tolo)

        print('Starting quantum simulation')
        
        result_simu = {}
        result_simu['data'] = np.zeros(nsteps, dtype=np.float64)
        if(Is_store_circuit):  result_simu['circuits'] = [ [] for _ in range(nsteps) ]

        for i in range(nsteps):
            print('Simulation step', i, 'of', nsteps)
            current_kraus_list = Kraus[i]
            print('Number of Kraus operators:', len(current_kraus_list))
            for kraus_op in current_kraus_list:
                for istate in range(n_inistate):
                    qc = self._create_circuit(kraus_op, statevec[istate], Isscale=False)
                    result = estimator.run([(qc, obs_q)]).result()
                    
                    if(Is_store_circuit):  result_simu['circuits'][i].append(qc)
                    result_simu['data'][i] += result[0].data.evs * prob[istate]

        return result_simu

    def qc_simulation_vecdens(
        self,
        time_arr: List[float],
        shots: int = 1024,
        backend: Any = AerSimulator(),
        Gprop: Optional[List[np.ndarray]] = None,
        **kwargs: Any
    ) -> np.ndarray:
        """
        Perform quantum simulation using the vectorized density matrix representation.

        This method simulates the quantum system dynamics over a series of time steps by constructing circuits
        based on the vectorized density matrix representation, performing measurements, and processing the results.

        Args:
            time_arr (List[float]): List of time steps for simulation.
            shots (int, optional): Number of measurement shots. Defaults to 1024.
            backend (Any, optional): Quantum simulation backend. Defaults to AerSimulator().
            Gprop (Optional[List[np.ndarray]], optional): Propagator matrix (or list of matrices) for simulation.
                If None, it will be calculated. Defaults to None.
            **kwargs: Additional keyword arguments for propagator calculation.

        Returns:
            Dict[str, Any]: A dictionary containing the quantum simulation results, including:
                - 'data' (np.ndarray): Array of shape (nsteps, n_bitstr), storing the processed probability amplitudes 
                (i.e., normalized square root of the measured probabilities, scaled by norm factors) at each time step.
                - 'circuits' (List[QuantumCircuit]): List of generated quantum circuits for each time step.
                - 'norm' (List[float]): List of norm correction factors used in post-processing at each time step.
        """
        if Gprop is None:
            Gprop = self.Gt_matrix_expo(time_arr, **kwargs)

        nsteps = len(time_arr)

        if self.count_str is None:
            print("Error: count_str is not assigned")

        n_bitstr = len(self.count_str)
        statevec, norm0 = self.init_statevec_vecdens()
        
        result = {}
        result['data'] = np.zeros((nsteps, n_bitstr), dtype=np.float64)
        result['circuits'] = []
        result['norm'] = []
        
        for i in range(nsteps):
            if i % 100 == 0:
                print('Quantum simulation step', i)
            Gt = Gprop[i]
            
            #create the circuit
            circuit, norm = self._create_circuit(Gt, statevec, Isscale=True)
            circuit.measure(range(self.Nqb + 1), range(self.Nqb + 1))
            if self.dilation_method == 'SVD-Walsh':
                circuit = transpile(circuit, backend)
            
            #store the circuits and norm to the result
            result['circuits'].append(circuit)
            result['norm'].append(norm)
            
            counts = backend.run(circuit, shots=shots).result().get_counts()
            for j in range(n_bitstr):
                bitstr = self.count_str[j]
                if bitstr in counts:
                    result['data'][i, j] = np.sqrt(counts[bitstr] / shots) * norm * norm0
                else:
                    print('At time', i, 'with shots =', shots, "no counts for", bitstr)

        return result

    def _create_circuit(
        self,
        array: np.ndarray,
        statevec: Union[np.ndarray, List[np.ndarray]],
        Isscale: bool = True
    ) -> QuantumCircuit:
        """
        Construct and return the quantum circuit.

        This method wraps the call to the dilation circuit construction function.

        Args:
            array (np.ndarray): Array used for circuit construction (e.g., propagator or Kraus operator).
            statevec (Union[np.ndarray, List[np.ndarray]]): State vector(s) to be used in the circuit.
            Isscale (bool, optional): Flag indicating whether scaling should be applied. Defaults to True.

        Returns:
            QuantumCircuit: The constructed quantum circuit.
        """
        return dc.construct_circuit(self.Nqb, array, statevec, method=self.dilation_method, Isscale=Isscale)



import numpy as np
import scipy.linalg as LA

from qiskit import QuantumRegister, ClassicalRegister, QuantumCircuit
from qiskit.quantum_info.operators import Operator

from . import walsh_gray_optimization as wo


def scale_array(array,scale=1.1):
    """
    renormalize an array to make sure it is a contraction
    """
    # Normalization factor, divide by martix's norm to ensure contraction
    # may divide a larger scaling factor controlled by the number (scale)
    norm = LA.norm(array,2)*scale
    array_new = array/norm
    
    return array_new, norm

def dilate_Sz_Nagy(array):
    """
    dilate the non-unitary array (should be a contraction) to a unitary matrix
    array: ndarray of N*N
    """
    ident = np.eye(array.shape[0])
      
    # Calculate the conjugate transpose of the G propagator
    fcon = (array.conjugate()).T
      
    # Calculate the defect matrix for dilation
    fdef = LA.sqrtm(ident - np.dot(fcon, array))
      
    # Calculate the defect matrix for the conjugate of the G propagator
    fcondef = LA.sqrtm(ident - np.dot(array, fcon))
      
    # Dilate the G propagator to create a unitary operator
    array_dilated = np.block([[array, fcondef], [fdef, -fcon]])
    
    return array_dilated
          
#generate the quantum gate matrices for SVD-dilation
def dilate_SVD(array):
    """
    dilate the non-unitary array to unitary matrices using SVD technique by Schlimgen et al
    (Phys. Rev. A 2022, 106, 022414.)
    
    array: ndarray of N*N (should be a contraction)
    """
    
    #get the dimension of the array
    Nvec = array.shape[0]
    
    #Performing SVD to the array
    U1,S1,V1 = LA.svd(array)

    Mzero = np.zeros((Nvec,Nvec),dtype=np.complex128)
    fk=np.zeros(2*Nvec,dtype=np.float64)
    
    Sig_p = np.zeros(Nvec,dtype=np.complex128)
    Sig_m = np.zeros(Nvec,dtype=np.complex128)
    for i in range(len(S1)):
        
        Sig_p[i] = S1[i]+1j*np.sqrt((1-S1[i]**2))
        Sig_m[i] = S1[i]-1j*np.sqrt((1-S1[i]**2))
    
        #here U_Sigma = e^{i f_k}
        fk[i] = (-1j*np.log(Sig_p[i])).real
        fk[Nvec+i] = (-1j*np.log(Sig_m[i])).real
    
    SG0 = np.block([[np.diag(Sig_p),Mzero],\
          [Mzero,np.diag(Sig_m)]])

    return U1, V1, SG0, fk


def cons_SVD_cirq(Nqb, array, ini_vec, Iswalsh = True):
    """
    Construct the SVD-dilation circuit
    Nqb: number of qubits
    array: the non-unitary array (should be a contraction)
    ini_vec: initial qubit state vector (in the dilation space)
    Iswalsh: If True, then combine with Walsh operator representation to reduce the circuit depth
    """
    
    qc = QuantumCircuit(Nqb,Nqb)
    qc.initialize(ini_vec,range(0,Nqb))
    
    U_matu,U_matv,S_mat0,fk = dilate_SVD(array)
    
    qc.append(Operator(U_matv),range(0,Nqb-1))
    qc.h(Nqb-1)

    if(Iswalsh):
        #the walsh coeff
        arr_a = wo.walsh_coef(fk,Nqb)
        
        Ulist_diag0 = wo.cirq_list_walsh(arr_a,Nqb,1E-5)
        Ulist_diag = wo.optimize(Ulist_diag0)
        qc_diag = wo.cirq_from_U(Ulist_diag,Nqb)
        
        qc.append(qc_diag.to_gate(),range(Nqb))

    else:
        qc.append(Operator(S_mat0),range(Nqb))
      
    qc.append(Operator(U_matu),range(0,Nqb-1))
    qc.h(Nqb-1)

    return qc

def cons_SzNagy_cirq(Nqb, array, ini_vec):
    """
    Construct the Sz-Nagy-dilation circuit
    Nqb: number of qubits
    array: the non-unitary array (should be a contraction)
    ini_vec: initial qubit state vector (in the dilation space)
    """
    
    qr = QuantumRegister(Nqb)  # Create a quantum register
    cr = ClassicalRegister(Nqb)  # Create a classical register to store measurement results
    qc = QuantumCircuit(qr, cr)  # Combine the quantum and classical registers to create the quantum circuit
    
    # Initialize the quantum circuit with the initial state
    qc.initialize(ini_vec, qr)
    
    # Create a custom unitary operator with the dilated propagator
    U_dil = dilate_Sz_Nagy(array)
    U_dil_op = Operator(U_dil)
    
    # Apply the unitary operator to the quantum circuit's qubits
    qc.unitary(U_dil_op, qr)

    return qc

def construct_circuit(Nqb, array,statevec,method = 'Sz-Nagy',Isscale = True):
    """
    construct the quantum circuit
    
    method: specify the dilation method, can be 'Sz-Nagy' or 'SVD' or 'SVD-Walsh'
        'Sz-Nagy': using Sz-Nagy method do dilation
        'SVD': using SVD-dilation method 
        'SVD-Walsh': SVD-dilation combine with Walsh operator representation to reduce the circuit depth
    
    Nqb: number of qubits in the original space, should match the dimension of array (N=2^Nqb)
    array: ndarray of N*N
    statevec: ndarray of N*1 (initial statevector)
    
    Isscale: 
        if Ture, renormalize the array to make sure it is a contraction
        if False, do not renormalize. 
    """   
    
    #the number of qubits for the original space
    if(2**Nqb != array.shape[0]):
        print('error, array dimension not matched!')
    
    #state vector in the dilated space
    statevec_dil = np.concatenate((statevec, np.zeros_like(statevec)))
    
    #first scale the array to make it a contraction
    if(Isscale):
        array_new, normfac = scale_array(array)
    else:
        array_new = array
    
    if(method == 'Sz-Nagy'):
        qc = cons_SzNagy_cirq(Nqb+1, array_new, statevec_dil)
    elif(method == 'SVD'):
        Iswalsh = False
        qc = cons_SVD_cirq(Nqb+1, array_new, statevec_dil, Iswalsh)
    elif(method == 'SVD-Walsh'):
        Iswalsh = True
        qc = cons_SVD_cirq(Nqb+1, array_new, statevec_dil, Iswalsh)
    else:
        print('method error in construct circuit')
    
    if(Isscale):
        return qc,normfac
    else:
        return qc





import logging
import logging.config
import os
from pathlib import Path

import tomli


def setup_logging() -> None:
    """Configures logging for the application.

    Reads the logging configuration from the 'pyproject.toml' file
    and applies it using `logging.config.dictConfig`.

    The log level can be overridden by setting the
    'QFLUX_LOG_LEVEL' environment variable. If the environment
    variable is not set or set to an invalid level, it defaults to 'INFO'.

    Raises:
        FileNotFoundError: If 'pyproject.toml' is not found.
        tomli.TOMLDecodeError: If 'pyproject.toml' is not a valid TOML file.
        KeyError: If the 'tool.logging' section is missing in 'pyproject.toml'.
    """
    # Construct the path to pyproject.toml, assuming it's 3 levels up from this file.
    pyproject_path = Path(__file__).parents[3] / "pyproject.toml"

    try:
        with open(pyproject_path, "rb") as f:
            config = tomli.load(f)
    except FileNotFoundError as e:
        raise FileNotFoundError(f"pyproject.toml not found at {pyproject_path}: {e}") from e
    except tomli.TOMLDecodeError as e:
        raise tomli.TOMLDecodeError(f"Failed to decode pyproject.toml: {e}") from e

    # Determine the log level:
    # 1. Check for QFLUX_LOG_LEVEL environment variable.
    # 2. Default to 'INFO' if not set or invalid.
    log_level_env = os.getenv("QFLUX_LOG_LEVEL", "INFO").upper()

    valid_levels = {"DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"}
    if log_level_env not in valid_levels:
        print(f"Invalid log level '{log_level_env}' from environment, defaulting to INFO")
        log_level = "INFO"
    else:
        log_level = log_level_env

    try:
        # Extract logging configuration from pyproject.toml
        logging_config = config["tool"]["logging"]
    except KeyError as e:
        raise KeyError(f"'tool.logging' section not found in pyproject.toml: {e}") from e

    # Override the log level specified in pyproject.toml with the determined log level.
    logging_config["loggers"]["qflux"]["level"] = log_level

    # Configure logging using the dictionary configuration.
    logging.config.dictConfig(logging_config)


# Get the logger for the 'qflux' application.
logger = logging.getLogger("qflux")



"""
Example Parameters class as illustrated in Part_I_JCE.ipynb.

"""

import numpy as np

class Params:
    def __init__(self):
        # ==== Spin-Boson Model parameters ====
        self.GAMMA_DA = 1          # diabatic coupling
        self.EPSILON  = 1
        self.BETA     = 5          # inverse temperature beta = 1 / (k_B * T)
        self.XI       = 0.1
        self.OMEGA_C  = 2

        # Spin-up and spin-down states
        self.spin_up   = np.array([1.0, 0.0], dtype=np.float64)
        self.spin_down = np.array([0.0, 1.0], dtype=np.float64)

        # ==== General constants for simulation ====
        self.TIME_STEPS = 500                    # number of time steps
        self.au2ps      = 0.00002418884254       # as -> a.u. conversion
        self.timeau     = 12.409275
        self.DT         = 20 * self.au2ps * self.timeau  # time step in au

        self.FINAL_TIME = self.TIME_STEPS * self.DT
        self.DOF_E      = 2                      # number of electronic states
        self.DOF_E_SQ   = self.DOF_E * self.DOF_E

        # ==== Simulation parameters for TT-TFD ====
        self.DOF_N      = 50     # number of nuclear DOF
        self.OMEGA_MAX  = 10

        # TT constants
        self.eps         = 1e-12         # tt approx error
        self.dim         = self.DOF_N    # number of coords
        self.occ         = 10            # max occupation number
        self.MAX_TT_RANK = 10

        # ==== Simulation parameters for GQME ====
        self.MEM_TIME          = self.DT * self.TIME_STEPS
        self.HBAR              = 1
        self.MAX_ITERS         = 30
        self.CONVERGENCE_PARAM = 10.0**(-10.0)

        # ==== Parameter string for output files ====
        self.PARAM_STR  = "_Spin-Boson_Ohmic_TT-TFD_b%sG%s_e%s_" % (
            self.BETA, self.GAMMA_DA, self.EPSILON
        )
        self.PARAM_STR += "xi%swc%s_wmax%s_dofn%s" % (
            self.XI, self.OMEGA_C, self.OMEGA_MAX, self.DOF_N
        )

        # ==== Pauli matrices ====
        self.X = np.array([[0, 1], [1, 0]], dtype=np.complex128)
        self.Y = np.array([[0, -1j], [1j, 0]], dtype=np.complex128)
        self.Z = np.array([[1, 0], [0, -1]], dtype=np.complex128)
        self.I = np.eye(2, dtype=np.complex128)



import requests
import os
from pathlib import Path

def load_dataset(ds_path: Path) -> None:
    """Loads a dataset from the specified path.

    This function is intended to load a dataset from the given file path.
    Currently, it is not implemented and always raises a NotImplementedError.

    Args:
        ds_path: The path to the dataset file.

    Raises:
        NotImplementedError: Always raised as the function is not yet implemented.
    """
    raise NotImplementedError("Not implemented")

def download_github_directory(repo_owner, repo_name, dir_path, local_dir):

    """
    Recursively download a directory from a GitHub repository
    """

    #-- GitHub API URL for the directory contents
    url = f"https://api.github.com/repos/{repo_owner}/{repo_name}/contents/{dir_path}"
    response = requests.get(url)
    
    if response.status_code != 200:
        print(f"Error: {response.status_code} - {response.json().get('message', 'Unknown error')}")
        return
    
    contents = response.json()
    
    #-- Create the local directory if it doesn't exist
    os.makedirs(local_dir, exist_ok=True)
    
    for item in contents:

        if item['type'] == 'file':
            # Download the file
            file_url = item['download_url']
            file_name = item['name']
            file_path = os.path.join(local_dir, file_name)
            
            print(f"Downloading: {file_path}")
            file_response = requests.get(file_url)
            with open(file_path, 'wb') as f:
                f.write(file_response.content)
                
        elif item['type'] == 'dir':
            # Recursively download subdirectory
            subdir_path = item['path']
            subdir_local = os.path.join(local_dir, item['name'])
            print(f"===> Directory: {subdir_path}")
            download_github_directory(repo_owner, repo_name, subdir_path, subdir_local)



"""
Define some Parameter and Constants
"""

import numpy as np

# Pauli matrices
X = np.array([[0, 1], [1, 0]], dtype=np.complex128)
Y = np.array([[0, -1j], [1j, 0]], dtype=np.complex128)
Z = np.array([[1, 0], [0, -1]], dtype=np.complex128)
I = np.eye(2, dtype=np.complex128)


# sigma+ and sigma-
sigmap = 0.5*(X+1j*Y)
sigmam = 0.5*(X-1j*Y)

# Spin-up and spin-down states
spin_up = np.array([1.0, 0.0], dtype=np.float64)
spin_down = np.array([0.0, 1.0], dtype=np.float64)


#some constant to convert the units 
au2fs    = 2.418884254E-2
au2cm    = 219474.63068
au2joule = 4.35974381E-18
bolz     = 1.3806503E-23
au2ev    = 27.2114
hbar     = 1.0